package uk.ac.starlink.table;

import java.io.IOException;
import java.util.ArrayList;

/**
 * Wrapper table which allows conversion of cells in some columns.
 * Value {@link Converter}s can be installed on a per-column basis
 * to modify the data (and metadata) in that column.
 *
 * @author    Mark Taylor
 * @since     16 Sep 2008
 */
public class ConvertColumnStarTable extends WrapperStarTable {

    private final ArrayList converterList_;

    /**
     * Constructor.
     *
     * @param  base table
     */
    public ConvertColumnStarTable( StarTable base ) {
        super( base );
        converterList_ = new ArrayList();
    }

    /**
     * Sets the converter to use for a given column.
     *
     * @param  icol  column index
     * @param  converter   new converter to affect cells in that column
     */
    public void setConverter( int icol, Converter converter ) {
        converterList_.ensureCapacity( icol + 1 );
        converterList_.set( icol, converter );
    }

    /**
     * Returns the converter affecting a given column.
     * If no converter has been installed (straight pass-through of base
     * table data and metadata) then null is returned.
     *
     * @param  icol  column index
     * @return    converter, or null
     */
    public Converter getConverter( int icol ) {
        return icol < converterList_.size()
             ? (Converter) converterList_.get( icol )
             : null;
    }

    public ColumnInfo getColumnInfo( int icol ) {
        Converter conv = getConverter( icol );
        return conv == null ? super.getColumnInfo( icol )
                            : conv.getColumnInfo();
    }

    public Object getCell( long irow, int icol ) throws IOException {
        Converter conv = getConverter( icol );
        Object cell = super.getCell( irow, icol );
        return conv == null ? cell
                            : conv.convert( cell );
    }

    public Object[] getRow( long irow ) throws IOException {
        Object[] row = super.getRow( irow );
        for ( int icol = 0; icol < row.length; icol++ ) {
            Converter conv = getConverter( icol );
            if ( conv != null ) {
                row[ icol ] = conv.convert( row[ icol ] );
            }
        }
        return row;
    }

    public RowSequence getRowSequence() throws IOException {
        final int ncol = getColumnCount();
        final Converter[] convs = new Converter[ ncol ];
        for ( int icol = 0; icol < ncol; icol++ ) {
            convs[ icol ] = getConverter( icol );
        }
        return new WrapperRowSequence( super.getRowSequence() ) {
            public Object getCell( int icol ) throws IOException {
                Object cell = super.getCell( icol );
                return convs[ icol ] == null
                     ? cell
                     : convs[ icol ].convert( cell );
            }
            public Object[] getRow() throws IOException {
                Object[] row = super.getRow();
                for ( int icol = 0; icol < ncol; icol++ ) {
                    if ( convs[ icol ] != null ) {
                        row[ icol ] = convs[ icol ].convert( row[ icol ] );
                    }
                }
                return row;
            }
        };
    }

    /**
     * Defines data and metadata conversion to be applied to a single column.
     */
    public interface Converter {

        /**
         * Converts a value in the base table to the one returned by the
         * converted table.
         *
         * @param  value  base value
         * @return  converted value
         */
        Object convert( Object value );

        /**
         * Returns the metadata for the converted column.
         *
         * @return   column info
         */
        ColumnInfo getColumnInfo();
    }
}
